<?php

/**
 *
 * $Id: class.Migrator_20030910_02.php,v 1.1 2005/05/16 22:25:23 dgorski Exp $
 *
 * This file is part of coWiki. coWiki is free software under the terms of
 * the GNU General Public License (GPL). Read the LICENSE file. If you did
 * not receive a copy of the license and are not able to obtain it through
 * the internet, please send a note to <license@cowiki.org> so we can mail
 * you a copy immediately.
 *
 * @package     htdocs
 * @subpackage  setup
 * @access      public
 *
 * @author      Daniel T. Gorski, <daniel.gorski@develnet.org>
 * @copyright   (C) Daniel T. Gorski, {@link http://www.develnet.org}
 * @license     http://www.gnu.org/licenses/gpl.html
 * @version     $Revision: 1.1 $
 *
 */

/**
 * If required, this Migrator creates a database connector user and the
 * basic database scheme.
 *
 * @package     htdocs
 * @access      public
 *
 * @author      Daniel T. Gorski, <daniel.gorski@develnet.org>
 * @since       coWiki 0.4.0
 */
class Migrator_20030910_02 implements MigratorInterface {

    private
        $State = null;

    // --------------------------------------------------------------------

    public function __construct($StateHolder) {
        $this->State = $StateHolder;
    }

    // --------------------------------------------------------------------

    /**
     *  Get a short description of what this Migrator is going to do
     */
    public function getDescription() {
        return 'Initial creation of the MySQL document database schema
                (if necessary).';
    }

    // --------------------------------------------------------------------

    /**
     *  Check if this Migrator has to be performed, its task is:
     *
     *  Initial creation of the MySQL document database schema
     */
    public function isRequired() {

        // Get the resource DSN (database connection details)
        $UriInfo = new UriInfo($this->State->getDocumentStorageDSN());

        // Do we use MySQL? Check if this Migrator can/should generally
        // handle the requested connector scheme. If not, this Migrator is
        // "not required".
        switch ($UriInfo->get('scheme')) {
            case 'mysql':
            case 'mysql+innodb':
            case 'mysqli':
            case 'mysqli+innodb':
                break;

            default:
                // Not a known MySQL scheme, Migrator not required
                return false;
        }

        return $this->State->shallCreateNewDocumentDatabase();
    }

    // --------------------------------------------------------------------

    /**
     *  Perform the Migrator task
     */
    public function perform() {

        // Get the resource DSN (database connection details)
        $UriInfo = new UriInfo($this->State->getDocumentStorageDSN());

        $sUser = $UriInfo->get('user');
        $sPass = $UriInfo->get('pass');
        $sHost = $UriInfo->get('host');
        $sDB   = $UriInfo->get('basepath');

        // ---

        // Connect to database with _root_ (DBA) crendentials!
        $sResource = $this->State->getDocumentStorageRootDSN();
        $Storage = StorageFactory::getInstance()
                        ->create(new UriInfo($sResource));

        // ---

        $sPath = $this->State->getMiscPath();
        $sFile = '/database/mysql.sql';

        $sContent = @file_get_contents($sPath . $sFile);

        if (trim($sContent) == '') {
            throw new MigratorException(
                'The MySQL scheme file <tt>'.$sFile.'</tt> in the
                <tt>misc/</tt> directory was either empty, not readable
                or absent.'
            );
        }

        // Remove remarks
        $sContent = str_replace("\r", '', $sContent);
        $sContent = preg_replace('=^#.*\n=m', '', $sContent);

        // Split into single queries, semicolons must not appear within
        // queries!
        $aContent = preg_split(
                        '#;\s*\n#',
                        $sContent,
                        -1,
                        PREG_SPLIT_NO_EMPTY
                    );

        // ---

        // Begin transaction
        $Storage->begin();

        // Check if database already exists
        $aDB = $Storage->fetchRowAndFreeResult(
            $Storage->query('SHOW DATABASES LIKE "'.$sDB.'"')
        );

        // A database with this name was found. Check if it seems to
        // contain valid tables. Avoid accidents.
        if (isset($aDB[0])) {
            $Storage->useDatabase($sDB);

            // Check if database contains tables that are prefixed
            // with "cowiki_"
            $aTable = $Storage->fetchRowAndFreeResult(
                $Storage->query('SHOW TABLES LIKE "cowiki_%"')
            );

            if (!is_array($aTable)) {
                $Storage->rollback();

                throw new MigratorException(
                    'The selected database \''.$sDB.'\' is already there
                    and does not seem to be a valid '.COWIKI_NAME.'
                    database. Heck, what are you doing? Be careful,
                    you are logged on as database administrator!'
                );
            }
        }

        // No database with this name found, create it 
        if (!isset($aDB[0])) {
            $Storage->query('CREATE DATABASE '.$sDB);
            $Storage->useDatabase($sDB);

            // Create basic schema
            foreach ($aContent as $sQuery) {
                $Storage->query($sQuery);
            }
        }

        // End transaction
        $Storage->commit();

        // ---

        // Now we have a valid connection to the storage; with a valid
        // connector user and a valid database schema. This setting is
        // required for successive Migrators.
        $this->State->setHasValidDocStorageConnection(true);

        // Done
        $this->State->setShallCreateNewDocumentDatabase(false);
    }

} // of class

?>