<?php

/**
 *
 * $Id: class.FileOutputStream.php,v 1.16 2005/06/08 13:50:01 dgorski Exp $
 *
 * This file is part of coWiki. coWiki is free software under the terms of
 * the GNU General Public License (GPL). Read the LICENSE file. If you did
 * not receive a copy of the license and are not able to obtain it through
 * the internet, please send a note to <license@cowiki.org> so we can mail
 * you a copy immediately.
 *
 * @package     io
 * @subpackage  class
 * @access      public
 *
 * @author      Daniel T. Gorski, <daniel.gorski@develnet.org>
 * @copyright   (C) Daniel T. Gorski, {@link http://www.develnet.org}
 * @license     http://www.gnu.org/licenses/gpl.html
 * @version     $Revision: 1.16 $
 *
 */

/**
 * coWiki - File output stream class
 *
 * @package     io
 * @subpackage  class
 * @access      public
 *
 * @author      Daniel T. Gorski, <daniel.gorski@develnet.org>
 * @since       coWiki 0.3.0
 *
 * @todo        [D11N]  Complete documentation
 */
class FileOutputStream extends AbstractOutputStream {

    protected
        $rFile     = null,
        $bOpener   = false,
        $sLockType = '',
        $sLockFile = '';

    // --------------------------------------------------------------------

    /**
     * Class constructor
     *
     * @access  public
     * @param   object
     * @param   boolean
     * @param   string
     * @return  void
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.3.0
     *
     * @throws  FileNotWritableException
     *
     * @todo    [D11N]  Check the parameter type of "$mOut"
     */
    public function __construct($mOut, $bAppend = false, $sLockType = 'STD') {

        // Remember locking strategy ('STD', 'FILE', 'NONE')
        $this->sLockType = $sLockType;

        if (is_string($mOut)) {

            // If locking strategy is 'STD' (standard) and we are running
            // on Windows, we force the strategy to be 'FILE' which will
            // use a .lock file to emulate a lock.
            $bForceDotLock = strpos(strtolower(PHP_OS), 'win') === true;

            if ($this->sLockType == 'STD' && $bForceDotLock) {
                $this->sLockType = 'FILE';
            }

            // ---

            // {{{ DEBUG }}}
            Logger::io('Trying to open for writing "'.$mOut.'"');

            $sMode = $bAppend ? 'ab' : 'wb';

            if (!$this->open($mOut, $sMode)) {
                throw new FileNotWritableException();
            }

            // This object opened the resource
            $this->bOpener = true;
        }

        if (is_resource($mOut)) {
           $this->rFile = $mOut;
        }
    }

    // --------------------------------------------------------------------

    /**
     * Class destructor
     *
     * @access  public
     * @return  void
     *
     * @author  Niels Braczek, <nbraczek@bsds.de>
     * @since   coWiki 0.3.4
     *
     * @todo    [D11N]  Check the parameter type of "$mOut"
     */
    public function __destruct() {
        $this->close();
    }

    // --------------------------------------------------------------------

    /**
     * Open
     *
     * @access  public
     * @param   string Filename
     * @param   string Access mode
     * @return  boolean
     *
     * @author  Niels Braczek, <nbraczek@bsds.de>
     * @since   coWiki 0.3.4
     *
     * @todo    [D11N]  Check description
     */
    public function open($sFilename, $sMode) {

        if ($this->sLockType == 'FILE') {
            if (!$this->lock($sFilename, LOCK_EX)) {
                return false;
            }
        }

        if (!($this->rFile = @fopen($sFilename, $sMode))) {
            return false;
        }

        if ($this->sLockType != 'FILE') {
            if (!$this->lock($this->rFile, LOCK_EX)) {
                $this->close();
                return false;
            }
        }

        return true;
    }

    // --------------------------------------------------------------------

    /**
     * Write
     *
     * @access  public
     * @param   string
     * @return  void
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.3.0
     *
     * @throws  FileNotWritableException
     *
     * @todo    [D11N]  Check description
     */
    public function write($sOut) {
        if (is_resource($this->rFile)) {
            if (!fwrite($this->rFile, $sOut)) {
                throw new FileNotWritableException();
            }
        }
    }

    // --------------------------------------------------------------------

    /**
     * Flush
     *
     * @access  public
     * @return  void
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.3.0
     *
     * @throws  FileNotWritableException
     *
     * @todo    [D11N]  Check description
     */
    public function flush() {
        if (is_resource($this->rFile)) {
            if(!fflush($this->rFile)) {
                throw new FileNotWritableException();
            }
        }
    }

    // --------------------------------------------------------------------

    /**
     * Close
     *
     * @access  public
     * @return  void
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.3.0
     *
     * @todo    [D11N]  Check description
     */
    public function close() {
        if ($this->bOpener && is_resource($this->rFile)) {
            $this->lock($this->rFile, LOCK_UN);
            @fclose($this->rFile);
        }
        $this->rFile   = null;
        $this->bOpener = false;
    }

    // --------------------------------------------------------------------

    /**
     * Portable locking depending passed on vlaues to the constructor.
     * STD  - Standard flock().
     * FILE - Uses lockfile.
     * NONE - No locking is performed.
     * @see flock()
     *
     * @access  public
     * @param   mixed Filename if FILE, filehandle if STD,
     * @param   int   Lock mode
     * @return  boolean
     *
     * @author  Niels Braczek, <nbraczek@bsds.de>
     * @since   coWiki 0.3.4
     *
     * @todo    [D11N]  Check description
     */
    private function lock($mFile, $iOperation) {
        $bSuccess  = false;

        switch ($this->sLockType) {

            case 'NONE':
                // No locking
                $bSuccess = true;
                break;

            case 'FILE':
                // dot-locking
                if ($iOperation == LOCK_UN && !empty($this->sLockFile)) {

                    if (@unlink($this->sLockFile)) {
                        $this->sLockFile = '';
                        $bSuccess = true;
                    }

                } else {

                    $sLockFile = $mFile.'.lock';
                    if (function_exists('link')) {
                        if (@link($mFile, $sLockFile)) {
                            $this->sLockFile = $sLockFile;
                            $bSuccess = true;
                        }
                    } else {
                        if ($rLock = @fopen($sLockFile, 'w')) {
                            @fclose($rLock);
                            $this->sLockFile = $sLockFile;
                            $bSuccess = true;
                        }
                    }

                }
                break;

            case 'STD':
            default:  // this case includes null/empty value
                $bSuccess = @flock($mFile, $iOperation);
                break;
        }

        return $bSuccess;
    }

} // of class

?>