<?php

/**
 *
 * $Id: class.Config.php,v 1.14 2005/04/26 05:23:04 dgorski Exp $
 *
 * This file is part of coWiki. coWiki is free software under the terms of
 * the GNU General Public License (GPL). Read the LICENSE file. If you did
 * not receive a copy of the license and are not able to obtain it through
 * the internet, please send a note to <license@cowiki.org> so we can mail
 * you a copy immediately.
 *
 * @package     parse
 * @subpackage  class
 * @access      public
 *
 * @author      Daniel T. Gorski, <daniel.gorski@develnet.org>
 * @copyright   (C) Daniel T. Gorski, {@link http://www.develnet.org}
 * @license     http://www.gnu.org/licenses/gpl.html
 * @version     $Revision: 1.14 $
 *
 */

/**
 * Configuration file manager and accessor class. This class provides 
 * capabilities of reading and writing a configuration file. The comments
 * in the configuration won't we destroyed.
 *
 * @package     parse
 * @subpackage  class
 * @access      public
 *
 * @author      Daniel T. Gorski, <daniel.gorski@develnet.org>
 * @since       coWiki 0.4.0
 */
class Config extends Object {

    private
        $sFileName  = '',
        $sLeadTxt   = '',
        $aConf      = array();

    // --------------------------------------------------------------------

    /**
     * Parse an .ini style configuration file.
     *
     * @access  public
     * @param   string  The name (complete path) of the config file to read.
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     *
     * @throws  ConfigNotReadableException
     * @throws  ConfigSyntaxException
     */
    public function __construct($sFileName) {
        $this->sFileName = $sFileName;

        try {
            $In = new FileInputStream($sFileName);
            $sStr = $In->readAll();
            $In->close();

        } catch (IOException $ioe) {
            throw new ConfigNotReadableException(
                'Failed to read '.$sFileName
            );
        }

        // ----------------------------------------------------------------

        // Remove all possible \r in 
        $sStr = str_replace("\r", '', $sStr);

        // Tweak for better matching
        $sStr = "\n" . trim($sStr);

        // Get leading text
        $sStr = preg_replace_callback(
                    '#^(.*)\n\[#Usm',
                    array($this, 'extractLeadingText'),
                    $sStr,
                    1
                );

        // '[' was eaten up by last regex pattern, replace it
        $sStr = '[' . $sStr;

        // ---

        // Split by sections name. Underscores (_) are not allowed in
        // section names.
        $aSections = preg_split(
                        '#^\[(\.?[A-Z0-9]+)\]#Um',
                        $sStr,
                        -1,
                        PREG_SPLIT_DELIM_CAPTURE | PREG_SPLIT_NO_EMPTY
                     );

        // ---

        $aConf = array();

        // Iterate through section matches and extract descriptions and
        // key-value pairs.
        for ($i=0, $n=sizeof($aSections); $i<$n; $i += 2) {

            // Try to match .ini style section content with comments.
            // Adding lowercase characters or numbers will increase the 
            // matching time dramatically. Watch this pattern anyway,
            // it eats time.
            preg_match_all(
                '#(.*)\n\s*([0-9A-Z_]+)\s*=\s*"([^"]*)"[^\n]*#SUms',
                $aSections[$i+1]."\n",
                $aEntries,
                PREG_SET_ORDER
            );

            $aEntry = array();

            // Extract values
            for ($j=0, $m=sizeof($aEntries); $j<$m; $j++) {
                $aEntry[$j]['DESC'] = trim($aEntries[$j][1]);
                $aEntry[$j]['KEY']  = $aEntries[$j][2];
                $aEntry[$j]['VAL']  = $aEntries[$j][3];
            }

            $aConf[$aSections[$i]] = $aEntry;
        }

        $this->aConf = $aConf;
    }

    // --------------------------------------------------------------------

    private function extractLeadingText($aMatches) {
        if ($aMatches[0]) {
            $this->setLeadingText(trim(substr($aMatches[0], 0, -2)));
        }
        return '';
    }

    // --------------------------------------------------------------------

    /**
     * Set the top leading text of the configuration file. This is usually
     * an explanation how to handle values in the configuration file.
     *
     * @access  public
     * @param   string  The leading text of the configuration file.
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     */
    public function setLeadingText($sStr) {
        $this->sLeadTxt = trim($sStr);
    
    }

    // --------------------------------------------------------------------

    /**
     * Get the leading text of the current configuration file. This value
     * might be empty.
     *
     * @access  public
     * @return  string  The leading text of the configuration file.
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     */
    public function getLeadingText() {
        return $this->sLeadTxt;
    }

    // --------------------------------------------------------------------

    /**
     * Reset all configuration settings.
     *
     * @access  public
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     */
    public function clear() {
        $this->aConf = array();
    }

    // --------------------------------------------------------------------

    /**
     * Return the configuration settings as associative array
     * (key-value pairs).
     *
     * Configuration settings that has been defined as 'yes', 'true' or 'on'
     * in a configuration file, are returned as boolean <tt>true</tt>
     * values. Accordingly 'no', 'false' and 'off' are boolean
     * <tt>false</tt>.
     *
     * @access  public
     * @return  array   Associative array with key-value pairs.
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     */
    public function &getConfigAsArray() {
        $aArr = array();
        $aTrue  = array('yes', 'true', 'on');
        $aFalse = array('no', 'false', 'off');

        foreach($this->aConf as $k => $v) {
            for ($i=0, $n=sizeof($v); $i<$n; $i++) {
                $sKey = $k . '_' . $v[$i]['KEY'];

                // Check for boolean value
                if (in_array(strtolower($v[$i]['VAL']), $aTrue)) {
                    $aArr[$sKey] = true;
                } else if (in_array(strtolower($v[$i]['VAL']), $aFalse)) {
                    $aArr[$sKey] = false;
                } else {
                    $aArr[$sKey] = $v[$i]['VAL'];
                }
            }
        }

        return $aArr;
    }

    // --------------------------------------------------------------------

    /**
     * Return the configuration settings as string (complete file).
     *
     * @access  public
     * @return  string  String that represents the whole configuration file.
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     */
    public function getConfigAsString() {
        $sStr = $this->getLeadingText() . "\n";

        foreach($this->aConf as $k => $v) {

            $sStr .= "\n[" . $k . "]\n";
                
            for ($i=0, $n=sizeof($v); $i<$n; $i++) {
                if ($v[$i]['DESC'] != '') {
                   $sStr .=  "\n    " . $v[$i]['DESC'] . "\n";
                }

                $sStr .=  '    ' . $v[$i]['KEY'] . ' = ';
                $sStr .=  '"' . $v[$i]['VAL'] . '"';
                $sStr .=  "\n";
            }

            // Section separator
            $sStr .=  "\n; " . str_repeat('-', 73) . "\n";
        }

        return trim($sStr) . "\n";
    }

    // --------------------------------------------------------------------

    /**
     * Set a key-value pair in a configuration section. Sections are parts
     * of the configuration file enclosed by square brackets. Example:
     * [SECTIONNAME]. Section names may only contain alphanumeric
     * characters.
     *
     * @access  public
     * @param   string  Section name without surrounding square brackets.
     * @param   string  Key name in the given section.
     * @param   string  New value. If you want to set a pseudo boolean
     *                  value, you may pass 'on' or 'off' here or use the 
     *                  setSectionKeyValueBoolean() method instead.
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     *
     * @see     setSectionKeyValueBoolean
     */
    public function setSectionKeyValue($sSect, $sKey, $sVal) {

        // Prepare and get rid of illegal characters in the [SECTION] and
        // section KEY strings.
        $sSect = $this->cleanSectionString($sSect);
        $sKey  = $this->cleanKeyString($sKey);

        // Also disallow illegal chars in the value
        $sVal = str_replace('"', '', $sVal);

        // If a section does not exist, create an empty one
        if (!isset($this->aConf[$sSect])) {
            $this->aConf[$sSect] = array();
        }

        // Check if a KEY in a [SECTION] alredy exists, set its value then.
        $bSet = false;
        for ($i=0, $n=sizeof($this->aConf[$sSect]); $i<$n; $i++) {
            if ($this->aConf[$sSect][$i]['KEY'] == $sKey) {
                $this->aConf[$sSect][$i]['VAL'] = $sVal;
                $bSet = true;
            }
        }

        // ... if it doesn't exist create a new entry.
        if (!$bSet) {
            $this->aConf[$sSect][$n]['DESC'] = '';
            $this->aConf[$sSect][$n]['KEY']  = $sKey;
            $this->aConf[$sSect][$n]['VAL']  = $sVal;
        }
    }

    // --------------------------------------------------------------------

    /**
     * Get configuration value by section name and its key.
     *
     * @access  public
     * @param   string  Section name without surrounding square brackets.
     * @param   string  Key name in the given section.
     * @return  mixed   If the key has been found, its value is returned.
     *                  If no key was found <tt>null</tt> is returned.
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     */
    public function getSectionKeyValue($sSect, $sKey) {
        if (isset($this->aConf[$sSect])) {
            for ($i=0, $n=sizeof($this->aConf[$sSect]); $i<$n; $i++) {
                if ($this->aConf[$sSect][$i]['KEY'] == $sKey) {
                    return $this->aConf[$sSect][$i]['VAL'];
                }
            }
        }

        return null;
    }

    // --------------------------------------------------------------------

    /**
     * Get a boolean configuration value by section name and its key.
     *
     * @access  public
     * @param   string  Section name without surrounding square brackets.
     * @param   string  Key name in the given section.
     * @return  mixed   If the key has been found, and it is pseudo boolean
     +                  expression (like 'on', 'off', 'yes' or 'no'), a
     *                  boolean value (true/false) will be returned.
     *                  If no key was found <tt>null</tt> is returned.
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     */
    public function getSectionKeyValueBoolean($sSect, $sKey) {
        if ($this->getSectionKeyValue($sSect, $sKey) === null) {
            return null;
        }

        return getBoolean($this->getSectionKeyValue($sSect, $sKey));
    }

    // --------------------------------------------------------------------

    /**
     * Set a boolean configuration value by section name and its key.
     *
     * @access  public
     * @param   string  Section name without surrounding square brackets.
     * @param   string  Key name in the given section.
     * @param   string  Boolean <tt>true</tt> or <tt>false</tt> value.
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     *
     * @throws  IllegalArgumentException
     */
    public function setSectionKeyValueBoolean($sSect, $sKey, $bVal = false) {

        // Must be boolean value
        if (!is_bool($bVal)) {
            throw new IllegalArgumentException('Value is not boolean');
        }

        // Prepare and get rid of illegal characters in the [SECTION] and
        // section KEY strings.
        $sSect = $this->cleanSectionString($sSect);
        $sKey  = $this->cleanKeyString($sKey);

        // If a section does not exist, create an empty one
        if (!isset($this->aConf[$sSect])) {
            $this->aConf[$sSect] = array();
        }

        // Check if a KEY in a [SECTION] alredy exists, set its value then.
        $bSet = false;
        for ($i=0, $n=sizeof($this->aConf[$sSect]); $i<$n; $i++) {
            if ($this->aConf[$sSect][$i]['KEY'] == $sKey) {
                $this->aConf[$sSect][$i]['VAL'] = $bVal ? 'on' : 'off';
                $bSet = true;
            }
        }

        // ... if it doesn't exist create a new entry.
        if (!$bSet) {
            $this->aConf[$sSect][$n]['DESC'] = '';
            $this->aConf[$sSect][$n]['KEY']  = $sKey;
            $this->aConf[$sSect][$n]['VAL']  = $bVal ? 'on' : 'off';
        }
    }

    // --------------------------------------------------------------------

    /**
     * Save the configuration file.
     *
     * @access  public
     * @param   string The complete path and name of the configuration
     *                 file to save. If no parameter is given, the path
     *                 defined in constructor is used.
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     *
     * @throws  ConfigNotWritableException
     */
    public function store($sFileName = null) {

        $sFileName = ($sFileName !== null) ? $sFileName : $this->sFileName;
        $sContent = $this->getConfigAsString();
        
        // If target platform is Win32, save the config file data with CRLF.
        if (substr(PHP_OS, 0, 3) == 'WIN') {
            $sContent = str_replace("\n", "\r\n", $sContent);
        }

        try {
            $Out = new FileOutputStream($sFileName);
            $Out->write($sContent);
            $Out->close();

        } catch (IOException $e) {
            throw new ConfigNotWritableException(
                'Could not write configuration file '.basename($sFileName)
            );
        }
    }

    // --------------------------------------------------------------------

    /**
     * Clean and prepare the configuration [SECTION] string
     *
     * @access  private
     * @param   string   Configuration file section string
     * @return  string   String without illegal characters
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     */
    private function cleanSectionString($sSect) {

        // Allow only valid section and key names. Hidden sections are
        // prefixed with a dot.
        $bHiddenSect = $sSect{0} === '.';

        // Clean, do not allow illegal characters. Divider chars like
        // dots (.) or underscores (_) are not allowed in section names.
        $sSect = preg_replace('#[^A-Z0-9]#', '', strtoupper($sSect));

        if ($bHiddenSect) { $sSect = '.' . $sSect; }
    
        return $sSect;
    }

    // --------------------------------------------------------------------

    /**
     * Clean and prepare the configuration [SECTION] KEY string
     *
     * @access  private
     * @param   string   Configuration file section key string
     * @return  string   String without illegal characters
     *
     * @author  Daniel T. Gorski, <daniel.gorski@develnet.org>
     * @since   coWiki 0.4.0
     */
    private function cleanKeyString($sKey) {
        return preg_replace('#[^A-Z0-9_.]#', '', strtoupper($sKey));
    }

} // of class

/*

    Heute bin ich recht gut ausgeschlafen aufgestanden
    Meine Zaehne geputzt und auf die Strasse gestolpert
    Und am Tageslicht erkannt, dass diese Gegend nicht dafuer vorgesehen ist
    Dass freundliche Menschen sich in ihr gegenseitig gut verstehen

    1 Uhr 15, 1 Uhr 30, 1 Uhr 45
    Manche sagen dreiviertel 2 dazu
    2 Uhr, 2 Uhr 15, sich gegenseitig gut verstehen

    Gestern war es aehnlich, lies es ruhig nach
    Nachts ist es meist dunkel und hell wird es dann am Tag
    Nicht dass ich es wollte, nur ich wuesste wirklich nicht warum ich mich
    Dafuer ... interessieren sollte

*/

?>
